#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Clases de alto nivel de la aplicación. 

Aquí se encuentran recogidas las clases encargadas tanto del despliegue como del
control de la aplicación. Es el "pegamento" del client_daemon
"""

from common.config import CLIENT_PORT
import logging
import os

# Pyro configuration
import Pyro
Pyro.config.PYRO_PORT = CLIENT_PORT
import Pyro.core #@Reimport

from implementations import ClientDataManager
from implementations import ClientIncidenceManager
from implementations import ClientUIAssistant
from managers import DataManager
from managers import IncidenceManager
from data import PyroFactory

from collectors import PingCollector
from collectors import SyncCollector
from collectors import IncChecker
from collectors import DummyGenerator
from collectors import TimeSyncGenerator
from collectors import DelayedGenerator
from collectors import UpdateGenerator
from collectors import BackupGenerator
from common.collectors import CollectorManager

from config import LOG_FILE
from common.config import LOG_FORMAT
from utils import RemoteLogger


class App:

    """
    Clase principal del demonio de los centros.
    """

    def __init__ (self, server_ip, centre_code = None):

        ''' Inicializar el objeto según parámetros de la línea de órdenes
            (los que se especificasen cuando se llamó a L{vaguada_client_daemon}).

            Este método existe únicamente para poder lanzar clientes especificando
            su código y la IP del servidor dinámicamente, en la línea de órdenes.
            B{El código del centro ya no se utiliza en línea de órdenes.}

            @param server_ip: dirección IP o nombre de un servidor I{Vaguada}.
                              Ejemplo: C{'vaguada-server'}.
            @type server_ip: L{str}
            @param centre_code: código del centro Guadalinfo (un par de códigos
                                numéricos, de provincia y municipio).
                                Ejemplo: C{(14, 60)}.
            @type centre_code: L{tuple} '''

        self.centre_code = centre_code
        PyroFactory.set_server_ip(server_ip)


        # Para preparar el sistema de testeo y tal
        os.system('mkdir -p /usr/share/vaguada')
        os.system('mkdir -p /var/vaguada/info/data')
        os.system('mkdir -p /var/vaguada/info/static_config')
        os.system('mkdir -p /var/vaguada/metadata')

        #os.system('cp /usr/share/vaguada/bootmisc.sh /etc/init.d/')

    def initialize(self):
        """
        Inicializa la aplicación, realizando el despliegue de los objetos PYRO.

        Pasos de la inicialización:
            - Se inicializan los diferentes hilos de los que consta la
              aplicación.
            - Se inicializan las esctructuras de datos necesarias.
            - Se despliegan los objetos PYRO.
        """

        self.__configure_logger()
        logger = logging.getLogger()

        logger.info("/----------------- Starting client daemon...")
        RemoteLogger.get_instance().log_inc("Inicio de ejecución")

        self.__initialize_data()
        self.__initialize_collectors()
        Deployer().deploy()
        logger.info("\----------------- Client daemon finished...")
        RemoteLogger.get_instance().log_inc("Fin de ejecución")

    def __configure_logger(self):
        logger = logging.getLogger()
        logger.setLevel(logging.INFO) 
        handler = logging.FileHandler(LOG_FILE)
        handler.setFormatter(logging.Formatter(LOG_FORMAT))

        logger.addHandler(handler)

    def __initialize_data(self):
        logger = logging.getLogger()
        IncidenceManager.get_instance().check_incidences ()
        logger.info('Data initialized')

    def __initialize_collectors(self):
        data_manager = DataManager.get_instance()

        # Create collectors
        ping = PingCollector(data_manager, timeout = 600)
        # Hacemos ping para poner el demonio en estado conectado si procede
        ping.collect()

        sync = SyncCollector(data_manager, timeout = 7200)

        inc = IncChecker(data_manager, timeout = 300)

        collectors = [ping, sync, inc]

        # Incidences generators
        dummy = DummyGenerator(data_manager, timeout = 2)
        collectors.append(dummy)

        timegenerator = TimeSyncGenerator(data_manager, timeout = 1800)
        collectors.append(timegenerator)

        updategenerator = UpdateGenerator(data_manager, timeout = 7200)
        collectors.append(updategenerator)

        backupgenerator = BackupGenerator(data_manager, timeout = 3600)
        collectors.append(backupgenerator)

        delayedgenerator = DelayedGenerator(data_manager, timeout = 1800)
        collectors.append(delayedgenerator)

        # Initialize collectors
        collector_manager = CollectorManager.get_instance(data_manager,
                collectors)
        collector_manager.init_collectors()



class Deployer:
    """
    Clase encargada del despliegue de los objetos Pyro.
    """

    def deploy(self):
        """
        Inicializa el despliegue de los objetos PYRO.

        Este método es bloqueante, es decir, una vez todos los objetos PYRO
        estén creados, se entrará en una espera no activa, proporcionada por el sistema
        Pyro.
        """
        Pyro.core.initServer()
        Pyro.core.initClient()

        daemon = Pyro.core.Daemon()

        data_manager_uri = daemon.connect(ClientDataManager(), "IClientDataManager")
        print "IClientDataManager uri:", data_manager_uri
        incidence_manager_uri = daemon.connect(ClientIncidenceManager(), "IClientIncidenceManager")
        print "IClientIncidenceManager uri:", incidence_manager_uri
        ui_assistant_uri = daemon.connect(ClientUIAssistant(), "IUIAssistant")
        print "IUIAssistant uri:", ui_assistant_uri

        try:
            daemon.requestLoop()
        except:
            CollectorManager.get_instance().stop_collectors()
