#!/usr/bin/python
# -*- coding: utf-8 -*-

"""Clases pertenecientes a los recolectores de datos.

Las clases definidas aquí son recolectores de datos. Inyectan datos en el
DataManager.

Existe un caso especial de recolector que se encarga de recoger/enviar ficheros
de información, y de mantener el sistema sincronizado con el servidor.
"""
import os
import os.path
import logging
import traceback
import time
import sys
import shutil
import difflib
import md5

from Pyro.core import ProtocolError
from Pyro.core import ConnectionClosedError
from Pyro.errors import URIError

from data import PyroFactory
from utils import connection_safe
from utils import CentreCodeError
from utils import RemoteLogger
from common.collectors import Collector
from common.collectors import LockPool
from managers import IncidenceManager

from incidences import BadLogIncidence
from incidences import UpdateIncidence
from incidences import DummyIncidence
from incidences import TimeIncidence
from incidences import CentreCodeIncidence
from incidences import ParseIncidence 
from incidences import DelayedIncidence 

from parsing import parse_auth_log
from parsing import parse_ldap

from config import DATA_DIR, MOD_FLAG_FILE, CHECKSUM_FILE
from common.data import FileData
from common.data import ModCache
from common.data import ChecksumCache

class PingCollector(Collector):
    """Mantiene al sistema informado de la conexión con el servidor

    Esto lo hace conectándose cada cierto tiempo al servidor y ejecutando su
    método ping. 
    """

    def collect(self):
        self.logger.debug("PingCollector: Collecting pings ...")

        logger = logging.getLogger()
        try:
            sdm = PyroFactory.get_server_data_manager()
            sdm.ping(self.data_manager.centre_code)
            DataManager.get_instance().connected = True

        except ProtocolError, e:
            logger.warning("ProtocolError: disconnected")
            DataManager.get_instance().connected = False

        except (URIError, ConnectionClosedError):
            logger.warning("ConnectionClosedError: disconnected")
            DataManager.get_instance().connected = False

        except CentreCodeError, e:
            logger.error("CentreCodeError: %s" % str(e))
            incidence = CentreCodeIncidence(str(e))
            IncidenceManager.get_instance().new_unique_incidence(incidence)

        except Exception:
            tbl = traceback.format_exception(sys.exc_type, 
                    sys.exc_value,
                    sys.exc_traceback)
            logger.error(''.join(tbl))


class IncChecker(Collector):
    """ Es el encargado de comprobar cada cierto tiempo si las incidencias
    todavía siguen vigente.
    """

    def collect(self):
        self.logger.debug("IncChecker: Checking incidences...")
        inc_manager = IncidenceManager.get_instance()
        for incidence in inc_manager.incidences.values():
            res = incidence.is_active()
            self.logger.debug("\t%s. Active:%s" % (str(incidence.__class__), str(res)))
            if res == False:
                inc_manager.remove_incidence(incidence)



class SyncCollector(Collector):
    """Es el encargado de iniciar cada cierto tiempo la sincronización de datos
    entre el servidor y el centro. 
    """

    def collect(self):
        self.logger.debug("SyncCollector: Collecting syncs...")
        # Generate data files:
        if self.__parse_files():
            lock = LockPool.get_lock("info")
            lock.acquire()
            self.data_manager.synchronize()
            lock.release()

    collect = connection_safe(collect)


    def __parse_files(self):
        incidence = None
        try:
            parse_auth_log()
        except Exception, e:
            tbl = traceback.format_exception(sys.exc_type, 
                    sys.exc_value,
                    sys.exc_traceback)
            desc = "Fallo al procesar el fichero de sesiones"
            incidence = ParseIncidence(desc, ''.join(tbl))

        try:
            parse_ldap()
        except Exception, e:
            tbl = traceback.format_exception(sys.exc_type, 
                    sys.exc_value,
                    sys.exc_traceback)
            desc = "Fallo al procesar el fichero de usuarios"
            incidence = ParseIncidence(desc, ''.join(tbl))

        if incidence:
            IncidenceManager.get_instance().new_unique_incidence(incidence)
            return False
        return True




# INCIDENCE GENERATORS ###################################################

class DummyGenerator(Collector):
    """ Comprueba si en el  sistema existe el fichero /tmp/vaguada_dummy_error.
    Si es así, genera la incidencia única L{DummyIncidence}
    """

    def collect(self):
        self.logger.debug('DummyGenerator: Checking for incidence...')
        if os.path.exists('/tmp/vaguada_dummy_error'):
            incidence = DummyIncidence()
            IncidenceManager.get_instance().new_unique_incidence(incidence)



class UpdateGenerator(Collector):
    """ Encargado de comprobar si las actualizaciones del sistema se han
    realizado o no.

    Genera una incidencia del tipo L{UpdateIncidence} si el sistema no está
    actualizado.
    """

    def collect(self):
        self.logger.debug('UpdateGenerator: Checking for incidence...')
        incidence = UpdateIncidence()

        if incidence.is_active():
            IncidenceManager.get_instance().new_unique_incidence(incidence)

    collect = connection_safe(collect)


class BackupGenerator(Collector):
    """ Clase encargada de copiar los ficheros configuración catalogados como
    estáticos al árbol de ficheros y directorios de vaguada.

    Los ficheros copiados por este recolector forman parte del sistema VAGUADA,
    es decir, tiene su correspondiente entrada en el fichero de sumas de chequeo
    y modificación
    """

    def collect(self):
        self.logger.debug("BackupGenerator: Collecting...")
        sdm = PyroFactory.get_server_data_manager()

        server_md5 = sdm.get_static_conf_list_checksum()

        if hasattr(self, 'static_conf_list_checksum'):
            if server_md5 == self.static_conf_list_checksum:
                return 

        self.static_conf_list_checksum = server_md5

        self.__generate_files()


    def __generate_files(self):
        sdm = PyroFactory.get_server_data_manager()
        files = sdm.get_static_conf_list()

        for scfile in files:
            if os.path.exists(scfile):
                filename = 'static_conf' + scfile
                dest = DATA_DIR + os.sep + filename


                # File diff
                diff = 'diff'
                if os.path.exists(dest):
                    dest_content = open(dest).read()
                    scfile_content = open(scfile).read()
                    diff = ''.join(difflib.context_diff(scfile_content, dest_content))

                if diff != '':
                    lock = LockPool.get_lock("files")
                    lock.acquire()
                    # Guardamos el fichero
                    dirpath = os.sep.join(dest.split(os.sep)[:-1])
                    os.system("mkdir -p " + dirpath)
                    shutil.copy(scfile, dest)

                    # Damos de alta al fichero en el sistema
                    content = open(dest).read()

                    ModCache(MOD_FLAG_FILE).add(filename)
                    checksum = md5.new(content).hexdigest()
                    ChecksumCache(CHECKSUM_FILE)[filename] = checksum

                    lock.release()


    collect = connection_safe(collect)
        


class TimeSyncGenerator(Collector):
    """ Encargado de comprobar que la hora del centro esté sincronizada con el
    servidor.
    """

    def collect(self):
        self.logger.debug ('TimeSyncGenerator: checking clock...')
        sdm = PyroFactory.get_server_data_manager()
        servertime = sdm.get_time()
        localtime = time.time()

        if abs(servertime - localtime) >= 600:
            try:
                self.__set_time(servertime)
            except SetTimeError:
                incidence = TimeIncidence()
                IncidenceManager.get_instance().new_unique_incidence(incidence)

    collect = connection_safe(collect)


    def __set_time(self, newtime):
        gmtime = time.gmtime(newtime)

        (year, month, day, hour, min) = gmtime[:5]
        timecommand = "date -u -s \"%(year)s/%(month)s/%(day)s %(hour)s:%(min)s\""\
        % locals()
        if os.system(timecommand) != 0:
            raise SetTimeError

        RemoteLogger.get_instance().log_inc("Hora sincronizada")



class DelayedGenerator(Collector):
    """ Encargado de genera una incidencia si se detecta que no se envían datos
    desde hace mucho tiempo
    """

    def collect(self):
        self.logger.debug("DelayedGenerator: checking for incidences...")
        incidence = DelayedIncidence()

        if incidence.is_active():
            IncidenceManager.get_instance().new_unique_incidence(incidence)




class SetTimeError(Exception):
    pass

from managers import DataManager
