#!/usr/bin/python
# -*- coding: utf-8 -*-

""" Clases pertenecientes a las incidencias que ocurren en los centros.


Cada clase de este módulo representa a un tipo de incidencia. En cada una de
ellas se debe disponer de métodos para solucionarlas. Dichos métodos devuelven
True si se ha solucionado la incidencia o False en cualquier otro caso.
"""
import time
import traceback
import logging
import os
import os.path
import datetime

from sets import Set

from common.incidences import Incidence
from common.incidences import IncidenceData
from data import PyroFactory
from common.data import FileMetaDataList
from utils import connection_safe
from utils import CentreCodeError
from utils import RemoteLogger
from utils import connection_safe
from utils import parse_dpkg_call
from config import DATA_DIR


class ClientIncidence(Incidence):
    """ Clase abstracta que representa una incidencia en el cliente.
    """

    def __init__(self, name, description):
        # Create new incidence ident.
        ident = (DataManager.get_instance().centre_code, time.time())
        Incidence.__init__(self, ident, name, description)

        # Manual solve method is always the firts one.
        self.methods.append("Manual")


    def solve(self, method_pos, *params):

        if method_pos == 0:

            manual_solve_method = ''
            # (Es necesario comprobar que "params" no es vacío)
            if len (params) > 0:
                manual_solve_method = '\n\tDescripción de la solución adoptada: '
                manual_solve_method += params[0]

            return self.manual_solve(manual_solve_method)

        return self.solve_others(method_pos)


    def solve_others(self, pos):
        """ Método utilizado por las clases hijas para resolver las incidencias.

        Este método lo han de redefinir todas las clases que deriven de ésta, y
        es utilizado para resolver las incidencias de forma automática.

        @type pos: int
        @param pos: Posición que ocupa el método a utilizar para resolver en la
        lista de métodos de la incidencia
        """
        msg = "Pos: %s" % str(pos)
        raise NotImplementedError, msg

    def manual_solve(self, solve_desc):
        """ Método de resolución manual de la incidencia
        """

        self.manual_solve_desc = solve_desc
        return True



class DummyIncidence(ClientIncidence):
    """ Incidencia "tonta", para pruebas
    """

    def __init__(self):
        ClientIncidence.__init__(self, 'Dummy incidence', 
                'Dummy long description')

        self.methods.append("Resolver")
        self.methods.append("No Resolver")
        self.methods.append("Borrar el fichero")


    def solve_others(self, pos):
        # Method: "Resolver"
        if pos == 1:
            return self.__solve()

        # Method: "No Resolver"
        if pos == 2:
            return self.__no_solve()

        # Method: "Borrar el fichero"
        if pos == 3:
            return self.__delete_file()

        return False


    def is_active(self):
        return os.path.exists('/tmp/vaguada_dummy_error')


    def __delete_file(self):
        os.remove('/tmp/vaguada_dummy_error')
        return True


    def __solve(self):
        return True


    def __no_solve(self):
        return False



# VAGUADA INCIDENCES ######################################################

class BadLogIncidence(ClientIncidence):
    """ Incidencia que indica que un ficheror de log está corrupto.
    """

    def __init__(self, logfile):
        """
        @type logfile: str
        @param logfile: Ruta del fichero de log corrupto.
        """
        ClientIncidence.__init__(self, "Fichero corrupto", 
                            "El fichero \"%s\" se encuentra en mal estado." %\
                            logfile)

        self.type = "BadLogIncidence"
        self.logfile = logfile


    def is_active(self):
        pass



class UpdateIncidence(ClientIncidence):
    """ Indica que el sistema no ha sido actualizado a la última versión
    """

    package_list_checksum = ''

    def __init__(self):
        ClientIncidence.__init__(self, "Sistema no actualizado", 
                "")

        self.type = "UpdateIncidence"
        self.active = False


    def is_active(self):
        sdm = PyroFactory.get_server_data_manager()

        server_md5 = sdm.get_package_list_checksum()

        if server_md5 == UpdateIncidence.package_list_checksum:
            return self.active

        UpdateIncidence.package_list_checksum = server_md5 
        server_pl = sdm.get_package_list()
        local_pl = parse_dpkg_call()

        only_in_local = local_pl - server_pl
        missed_in_local = server_pl - local_pl

        self.__format_details(only_in_local, missed_in_local)

        if only_in_local or missed_in_local:
            self.active = True
            return True

        self.active = False
        return False
    is_active = connection_safe(is_active)



    def __format_details(self, only_in_local, missed_in_local):
        self.details = "%-20s\t%-20s\t%-20s\n" %\
                ("Paquete", "Versión instalada", "Versión requerida") 
        self.details += "-----------------------------------"
        self.details += "------------------------------\n\n"
                
        only_dict = self.__to_dict(only_in_local)
        missed_dict = self.__to_dict(missed_in_local)

        only_keys = only_dict.keys()
        missed_keys = missed_dict.keys()
        globalkeys = Set(only_keys).union(missed_keys)
        globalkeys = list(globalkeys)

        self.description = "El sistema no ha sido "
        self.description += "actualizado a la última versión"
        self.description += "\nPaquetes con diferencias: %s" % len(globalkeys) 
        globalkeys.sort()

        for key in globalkeys:
            
            local_version = None
            if only_dict.has_key(key):
                local_version = only_dict[key]

            server_version = None
            if missed_dict.has_key(key):
                server_version = missed_dict[key]

            if local_version or server_version:
                line = "%-20s" % key + "\t"

                if local_version:
                    line += "%-20s" % local_version
                else:
                    line += "%-20s" % "Ninguna"

                line += "\t"
                if server_version:
                    line += "%-20s" % server_version
                else:
                    line += "%-20s" % "Ninguna"
                self.details += line + "\n"



    def __to_dict(self, pkg_set):
        result = {}
        pkg_list = list(pkg_set)
        pkg_list.sort()
        for name, version in pkg_list:
            result[name] = version

        return result




class SyncIncidence(ClientIncidence):
    """ Indica que la información sobre usuarios y conexiones no se ha podido
    sincronizar.
    """

    def __init__(self, missed_in_local, only_in_local):
        msg = self.__prepare_msg(missed_in_local, only_in_local)

        ClientIncidence.__init__(self, "Envíos no actualizados",
                msg)

        self.type = "SyncIncidence"

    def __prepare_msg(self, mil, oil):
        msg = ''


        if len(oil) > 0:
            msg = 'Ficheros únicamente en el centro:\n'
            oil.sort()

            for ele in oil:
                msg += '\t' + ele.sum_up() + '\n'

        msg += '\nFicheros únicamente en el servidor:\n'

        if len(mil) > 0:
            mil.sort()
            for ele in mil:
                msg += '\t' + ele.sum_up() + '\n'

        return msg


    def is_active(self):
        sdm = PyroFactory.get_server_data_manager()

        cc = DataManager.get_instance().centre_code

        dispatch_list =  sdm.get_dispatch_list(cc)
        local_list = FileMetaDataList(DATA_DIR)
        missed_in_local = local_list.get_missed_files(dispatch_list)
        only_in_local = dispatch_list.get_missed_files(local_list)

        return (len(missed_in_local) != 0  or len(only_in_local) !=0)

    is_active = connection_safe(is_active)



class CentreCodeIncidence(ClientIncidence):
    """ Indica que el código del centro no es válido.
    """

    def __init__(self, desc):
        """
        @type desc: str
        @param desc: Descripción larga de la incidencia. Indica el motivo por el
        cual el código del centro no es válido (porque el centro no esté dado de
        alta, etc.)
        """
        ClientIncidence.__init__(self, 'Código INE incorrecto', desc)
        self.type = "CentreCodeIncidence"


    def is_active(self):
        centre_code = DataManager.get_instance().centre_code
        sdm = PyroFactory.get_server_data_manager() 

        try:
            sdm.ping(centre_code)
            return False

        except CentreCodeError:
            return True

        except:
            logging.getLogger().debug(''.join(traceback.format_stack()))

    is_active = connection_safe(is_active)


    def discard(self):
        del self



class TimeIncidence(ClientIncidence):
    """ Indica que no se puede sincronizar la hora del cliente.
    """
    def __init__(self):
        ClientIncidence.__init__(self, 'Hora incorrecta',
                "No se puede sincronizar con la del servidor.")
        self.type = 'TimeIncidence'


    def is_active(self):
        sdm = PyroFactory.get_server_data_manager()
        servertime = sdm.get_time()
        localtime = time.time()
        return (abs(servertime - localtime) >= 600)

    is_active = connection_safe(is_active)





class ParseIncidence(ClientIncidence):
    """ Indica que no se han podido generar bien los ficheros de sesión o de
    usuarios.
    """

    def __init__(self, desc, details):
        name = "Fallo de procesamiento de datos"
        ClientIncidence.__init__(self, name, desc)
        self.details = details
        self.type = "ParseIncidence"


    def is_active(self):
        from parsing import parse_auth_log, parse_ldap
        try:
            parse_auth_log()
            parse_ldap()
            return False

        except:
            return True


class DelayedIncidence(ClientIncidence):
    """ Indica que no se han enviado datos al servidor desde hace mucho tiempo
    """

    MAXTIME = 5
    """ Número máximo de días permitidos sin enviar
    """

    def __init__(self):
        name = "Retraso en envíos"
        ClientIncidence.__init__(self, name, '')
        self.type = "DelayedIncidence"


    def is_active(self):
        sdm = PyroFactory.get_server_data_manager()
        dm = DataManager.get_instance()

        last_dispatch_ord = sdm.get_last_dispatch_date(dm.centre_code)
        today =  datetime.date.today()

        if last_dispatch_ord:
            last_dispatch_date = datetime.date.fromordinal(last_dispatch_ord)
            diff_in_days = (today - last_dispatch_date).days

            if diff_in_days >= DelayedIncidence.MAXTIME:
                self.description = "No se envían datos desde el %s" %\
                        str(last_dispatch_date)
                return True
        else:
            self.description = "No se han recibido datos de este centro"
            return True

        return False
    is_active = connection_safe(is_active)



class SCModificatedIncidence(ClientIncidence):
    """ Indica que se ha modificado uno o varios ficheros de configuración
    estática
    """

    def __init__(self, wrong_file):
        name = "Configuración modificada"
        desc = "Se ha modificado el fichero de configuración \"%s\"" % \
                wrong_file

        ClientIncidence.__init__(self, name, desc) 
        self.type = "SCModificatedIncidence"


    def is_active(self):
        return True

from managers import DataManager
