#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
Objetos encargados de administrar los datos del cliente.

Este módulo contiene el núcleo del cliente. Es aquí donde se encuentra gran
parte de la lógica interna del demonio del cliente.

"""
import pickle
import time
import logging
import md5
import os
import os.path
import shutil

from common.incidences import IncidenceData
from common.incidences import Incidence
from common.data import FileData
from common.data import FileMetaDataList
from config import DATA_DIR
from data import PyroFactory
from utils import connection_safe
from utils import RemoteLogger
from utils import IncidenceDict
from utils import GuadalinfoId


class DataManager(object):
    """
    <<Singleton>> Clase encargada de la gestión de datos del cliente.

    La principal responsabilidad de esta clase es mantener en todo momento 
    información sobre el estado del centro: lista de incidencias, hora en el 
    centro, código, etc.

    Otra funcionalidad que debe implementar es el aviso al servidor de la
    resolución de incidencias.

    """
    instance = None

    def get_instance():
        if not DataManager.instance:
            DataManager.instance = DataManager()
        return DataManager.instance
    get_instance = staticmethod(get_instance)


    def __init__(self):
        self.__centre_code = self.__set_cc_from_file()
        self.__connected = False
        self.logger = logging.getLogger()
        self.remote_logger = RemoteLogger.get_instance()


    # ATTRIBUTES ##############################################################

    # connected ###############################################################

    def __get_connected(self):
        return self.__connected

    def __set_connected(self, value):
        old_connected = self.__connected
        self.__connected = value

        if old_connected == False and \
                value == True:
            rl = RemoteLogger.get_instance()
            rl.sync()
            rl.log_inc("<CONNECTION> Conexión")
            IncidenceManager.get_instance().notify_ninc()


    connected = property(__get_connected, __set_connected, None, 
            """
            @type: boolean

            Estado de la conexión. C{True} si el centro está conectado, C{False}
            si no lo está.
            """)


    # centre_code #############################################################

    def __get_centre_code(self):
        return self.__centre_code


    def __set_centre_code(self, value):
        self.__centre_code = value

    centre_code = property(__get_centre_code, __set_centre_code, None,
            """Código del centro.
            """)

    # METHODS #################################################################

    def synchronize(self):

        sdm = PyroFactory.get_server_data_manager()
        self.remote_logger = RemoteLogger.get_instance()

        # Start sync
        sdm.start_sync(self.centre_code)
        self.__set_version()
        self.remote_logger.log_sync("Iniciando sincronización de ficheros")

        # Obtain dispatch_list from server
        dispatch_list = sdm.get_dispatch_list(self.centre_code)
        self.logger.debug("Files in server: %s" % \
                str([e.filename for e in dispatch_list]))

        # Calculate diff
        local_list = FileMetaDataList(DATA_DIR)
        self.logger.debug("Files in local: %s" % \
                str([e.filename for e in local_list]))
        missed_in_local = local_list.get_missed_files(dispatch_list)
        only_in_local = dispatch_list.get_missed_files(local_list)

        self.logger.debug("missed_in_local: %s" % str(missed_in_local.sum_up()))
        self.logger.debug("only_in_local: %s" % str(only_in_local.sum_up()))

        # Set and get files
        self.__set_files(only_in_local)
        self.__get_files(missed_in_local)

        # Set static_conf files or create incidence
        self.__process_static_conf(local_list, dispatch_list)

        # Check for incidence 
        dispatch_list = sdm.get_dispatch_list(self.centre_code)
        new_local_list = FileMetaDataList(DATA_DIR)
        new_local_list = FileMetaDataList(DATA_DIR)
        new_missed_in_local = new_local_list.get_missed_files(dispatch_list)
        new_only_in_local = dispatch_list.get_missed_files(new_local_list)

        if  len(new_missed_in_local) != 0 or \
                len(new_only_in_local) != 0:
            # Create incidence
            self.logger.debug("Creating UpdateIncidence")
            incidence = SyncIncidence(new_missed_in_local, new_only_in_local)
            IncidenceManager.get_instance().new_unique_incidence(incidence)

        # End sync
        sdm.end_sync(self.centre_code)
        self.remote_logger.log_sync("Sincronización de ficheros finalizada\n")

    synchronize = connection_safe(synchronize)


    def __process_static_conf(self, local_list, dispatch_list):
        """ Se encarga de enviar y recibir los ficheros válidos de configuración
        estática.

        El envío se puede producir en ambos sentidos, atendiendo a lo siguiente:
        
        Cliente -> Servidor:

            - El fichero está sólo en el cliente.

            - Existen copias distintas del fichero en el servidor y en el
              cliente, la copia del cliente se ha modificado de forma
              controlada, se ha creado una incidencia para comunicarlo al
              administrador y éste la ha aprobado.

        Servidor -> Cliente:

            - Existen copias distintas del fichero en el servidor y en el
              cliente, y la copia del servidor se ha modificado de forma
              controlada.
        """
        sdm = PyroFactory.get_server_data_manager()
        dispatch_list = sdm.get_dispatch_list(self.centre_code)

        for local_scfile in local_list:
            if local_scfile.type == 'static_conf':
                try:
                    server_scfile = dispatch_list.get_file(local_scfile.ident)
                except:
                    server_scfile = None

                # El fichero está en el servidor
                if local_scfile.checksum != server_scfile.checksum:
                    if not local_scfile.modified:
                        self.__receive(server_scfile)
                        # Restauramos el fichero en el sistema
                        real_path =  DATA_DIR + os.sep + server_scfile.ident
                        try:
                            os.makedirs(os.sep.join(server_scfile.ident.split(os.sep)[1:-1]))
                        except:
                            pass
                        dest = os.sep 
                        dest += os.sep.join(server_scfile.ident.split(os.sep)[1:])
                        shutil.copy(real_path, dest)
                            
                            

    __process_static_conf = connection_safe(__process_static_conf)


    def __send(self, file_meta):
        """ Envía el fichero indicado al servidor"""

        server_data_manager = PyroFactory.get_server_data_manager()
        # Create file_data from file_meta
        file_data = FileData.new_from_meta(file_meta)
        # Transmit
        server_data_manager.set_data(self.centre_code, file_data)
        # file_data.write()
        file_meta.modified = False

    __send = connection_safe(__send)


    def __set_files(self, file_meta_list):
        """Envía una lista de ficheros meta al servidor.

        Para ello, utiliza el método "set_data" del objeto remoto
        L{IServerDataManager}.
        """
        self.logger.debug("Sending %s files" % str(file_meta_list.sum_up()))
        file_list = [i.ident for i in file_meta_list if i.type == 'data']
        self.remote_logger.log_sync("\tEnviando los ficheros: %s" %\
                str(file_list))

        for file_meta in file_meta_list:
            # file_meta.type == 'data':
            if file_meta.corrupted:
                # FIXME: New incidence
                pass
            else:
                self.__send(file_meta)

    __set_files = connection_safe(__set_files)


    def __receive(self, file_meta):
        """ Recibe un fichero del servidor
        """
        server_data_manager = PyroFactory.get_server_data_manager()
        file_data = server_data_manager.get_data(self.centre_code, file_meta)
        file_data.write()
        file_data.modified = False


    def __get_files(self, file_meta_list):
        """Copia la lista de ficheros indicada en file_meta_list al servidor.

       
       Para ello utiliza el método "get_data" del objeto remoto
        L{IServerDataManager}.
        """
        self.logger.debug("Getting %s files" % str(file_meta_list.sum_up()))
        file_list = [i.ident for i in file_meta_list if i.type == 'data']
        self.remote_logger.log_sync("\tRecuperando los ficheros: %s" %\
                str(file_list))

        for file_meta in file_meta_list:
            self.__receive(file_meta)

    __get_files = connection_safe(__get_files)


    def __set_cc_from_file(self):
        gid = GuadalinfoId()
        prov_code = int(gid['ID_PROVINCIA'])
        mun_code = int(gid['ID_LOCALIDAD'])
        return (prov_code, mun_code)


    def __set_version(self):

        if os.path.exists('/etc/lsb-release'):
            content = open('/etc/lsb-release').read()
        else:
            content = ''

        sdm = PyroFactory.get_server_data_manager()
        sdm.set_version(self.centre_code, content)

    __set_version = connection_safe(__set_version)



class IncidenceManager(object):
    """
    <<Singleton>> Clase encargada de la administración/solución de incidencias.
    """

    instance = None

    def get_instance():
        if not IncidenceManager.instance:
            IncidenceManager.instance = IncidenceManager()
        return IncidenceManager.instance

    get_instance = staticmethod(get_instance)

    def __init__(self):
        self.__incidence_dict = IncidenceDict()


    # ATTRIBUTES ####

    def __get_incidences(self):
        return self.__incidence_dict

    incidences = property(__get_incidences, None, None,
            """ Incidencias actuales del centro.

            @type: dict
            """)

    # METHODS

    def check_incidences(self):
        inc_checker = IncChecker(DataManager.get_instance())
        inc_checker.collect()

    def new_incidence(self, incidence):
        """ Añade una nueva incidencia al sistema.

        Además de añadir la incidencia, este método registra la nueva incidencia
        en el log remoto (en el servidor).

        @type incidence: L{Incidence}
        @param incidence: Incidencia que se quiere añadir al sistema.

        """
        # Add new incidence to incidence list
        self.__incidence_dict[incidence.ident] = incidence

        # Add notification to log
        msg = "NUEVA INCIDENCIA: >>>>>>>>>> %s" % str(incidence)
        RemoteLogger.get_instance().log_inc(msg)

        # Notify ninc to server
        self.notify_ninc()

    new_incidence = connection_safe(new_incidence)



    def new_unique_incidence(self, incidence):
        """ Añade una incidencia a la lista de incidencias sii no existe otra
        incidencia del mismo tipo.
        """
        for myincidence in self.__incidence_dict.values():
            if myincidence == incidence:
                return None

        self.new_incidence(incidence)

        
    def solve(self, incidence_data, method_pos, *params):
        """ Resuelve la incidencia indicada a través del método determinado.

        @type incidence_data: L{IncidenceData}
        @param incidence_data: Datos de la incidencia.

        @type method_pos: int
        @param method_pos: Posición del método en la lista de métodos válidos de
        la incidencia por el cual se quiere resolver.

        @rtype: boolean
        @return: True si se ha solucionado la incidencia o si ya está resuelta
        (no está en la lista de incidencias).
        """
        from utils import RemoteLogger
        try:
            incidence = self.__incidence_dict[incidence_data.ident]
        except KeyError:
            return True
        
        # Notify to log
        try:
            methodname = incidence.methods[method_pos]
        except:
            methodname = "<Inexistente>"

        # Solve
        result = incidence.solve(method_pos, *params) 
        if result:
            # Remove incidence
            del self.__incidence_dict[incidence.ident]

            # Add notification to log
            msg = "INCIDENCIA RESUELTA: **********%s\tMétodo seleccionado: %s\n%s" %\
                                        (str(incidence),
                                        incidence.methods[method_pos],
                                        incidence.manual_solve_desc)
            RemoteLogger.get_instance().log_inc(msg)

            self.notify_ninc()

        else:
            msg = "INCIDENCIA _NO_ RESUELTA ~~~~~~~~~~ %s \tMétodo seleccionado: %s\n" %\
                    (str(incidence), methodname)
            RemoteLogger.get_instance().log_inc(msg)

        return result

    solve = connection_safe(solve)


    def remove_incidence(self, incidence):
        # Remove
        del self.incidences[incidence.ident]

        # Add notification to log
        msg = "INCIDENCIA BORRADA: ########## %s" % str(incidence)
        from utils import RemoteLogger
        RemoteLogger.get_instance().log_inc(msg)

        self.notify_ninc()

    remove_incidence = connection_safe(remove_incidence)


    def get_incidences_data(self):
        result = []
        for incidence in self.__incidence_dict.values():
            ident = incidence.ident
            name  = incidence.name
            desc = incidence.description
            new_inc = IncidenceData(ident, name, desc)

            new_inc.methods = incidence.methods
            new_inc.type = incidence.type
            new_inc.details = incidence.details


            result.append(new_inc)

        return result


    def notify_ninc(self):
            # Notify to server
            data_manager = DataManager.get_instance()
            sim = PyroFactory.get_server_incidence_manager()
            ninc = len(self.__incidence_dict)
            sim.set_ninc(data_manager.centre_code, ninc)

    notify_ninc = connection_safe(notify_ninc)



from incidences import SyncIncidence
from collectors import IncChecker
