#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
Utilidades para el demonio de los centros.
"""

import traceback
import logging
import os
import time
import sys

from  sets import Set

from Pyro.core import ProtocolError
from Pyro.core import ConnectionClosedError
from Pyro.errors import URIError 

from common.utils import CentreCodeError
from common.utils import PersistentDict
from data import PyroFactory

from config import DISPATCH_LOG_CACHE, INCIDENCE_LOG_CACHE
from config import INCIDENCE_DICT_FILE
from config import GUADALINFO_CONF
from common.config import LOG_FORMAT


def connection_safe(method):
    """ Decorador de métodos que necesitan informar al L{DataManager} de un fallo
    en la conexión.

    Se debe usar con todos lo métodos que hagan uso de algún metodo remoto, para
    controlar la gestión de la conexión.
    """

    def new_method (self, *args, **kargs):
        from managers import DataManager
        logger = logging.getLogger()
        result = None
        try:
            if DataManager.get_instance().connected:
                result = method(self, *args, **kargs)
                DataManager.get_instance().connected = True

        except ProtocolError, e:
            logger.warning("ProtocolError: disconnected")
            DataManager.get_instance().connected = False

        except (ConnectionClosedError, URIError):
            logger.warning("ConnectionClosedError: disconnected")
            DataManager.get_instance().connected = False

        except CentreCodeError, e:
            logger.error("CentreCodeError: %s" % str(e))
            incidence = CentreCodeIncidence(str(e))
            IncidenceManager.get_instance().new_unique_incidence(incidence)

        except Exception, e:
            tbl = traceback.format_exception(sys.exc_type, 
                    sys.exc_value,
                    sys.exc_traceback)
            logger.error(''.join(tbl))

        return result

    # Set new method doc from old method (for epydoc)
    new_method.__doc__ = method.__doc__
    return new_method


class RemoteLogger(object):
    """ <<Singleton>> Encapsula un sistema de bitácora.
    
    Este sistemam  envía los mensajes al
    servidor y, si éste está desconectado, mantiene una copia local que será
    enviada al servidor tan pronto como se restablezca la conexión.
    """
    instance = None

    def __init__(self):
        self.__configure_loggers()


    def get_instance():
        if not RemoteLogger.instance:
            RemoteLogger.instance = RemoteLogger()
        return RemoteLogger.instance

    get_instance = staticmethod(get_instance)


    def send(self, type, message):
        """ Escribe un mensaje en el log del servidor o, si éste no está
        conectado, en un fichero local
        """
        #from managers import DataManager
        data_manager = DataManager.get_instance()

        if data_manager.connected:
            # Remote logger
            sdm = PyroFactory.get_server_data_manager()
            message = "(%s, %s) " % data_manager.centre_code + message
            try:
                if type == 'INC':
                    sdm.log_inc(data_manager.centre_code, message)
                if type == 'SYNC':
                    sdm.log_sync(data_manager.centre_code, message)
            except (ProtocolError, ConnectionClosedError):
                data_manager.connected = False
        else:
            logging.getLogger(type).info(message)

    
    def log_sync(self, msg):
        self.send('SYNC', msg)


    def log_inc(self, msg):
        self.send('INC', msg)


    def sync(self):
        """ Envía todos los logs locales al servidor.
        """
        sdm = PyroFactory.get_server_data_manager()
        #from managers import DataManager
        data_manager = DataManager.get_instance()

        if data_manager.connected:
            filefunclist = [(DISPATCH_LOG_CACHE, "SYNC"),
                            (INCIDENCE_LOG_CACHE, "INC")]

            for filepath, type in filefunclist:
                logfile = open(filepath, 'r')
                lines = logfile.readlines()
                logfile.close()

                logfile = open(filepath, 'w')
                for line in lines:
                    line = ' || ' + line
                    line = line.strip()
                    try:
                        if type == 'SYNC': 
                            sdm.log_sync(data_manager.centre_code, line)
                        elif type == 'INC':
                            sdm.log_inc(data_manager.centre_code, line)

                    except (ProtocolError, ConnectionClosedError):
                        logfile.write(line)
                    except CentreCodeError:
                        logfile.write(line)

                logfile.close()


    def __configure_loggers(self):
        """ Configura los dos loggers que se usarán en local.

        Estos loggers son "INC" y "SYNC", y serán usado para la bitácora de
        incidencias y para la de envío de ficheros.
        """
        # INCIDENCE LOGGER
        inc_logger = logging.getLogger('INC')
        inc_logger.setLevel(logging.INFO) 

        handler = logging.FileHandler(INCIDENCE_LOG_CACHE)
        handler.setFormatter(logging.Formatter(LOG_FORMAT))

        inc_logger.addHandler(handler)

        # DISPATCH LOGGER
        sync_logger = logging.getLogger('SYNC')
        sync_logger.setLevel(logging.INFO)

        handler = logging.FileHandler(DISPATCH_LOG_CACHE)
        handler.setFormatter(logging.Formatter(LOG_FORMAT))

        sync_logger.addHandler(handler)



class IncidenceDict(PersistentDict):
    def __init__(self):
        PersistentDict.__init__(self, INCIDENCE_DICT_FILE)


class GuadalinfoId(dict):

    def __init__(self):
        dict.__init__(self)
        self.__parse_file()


    def __parse_file(self):
        
        try:
            fileobj = open(GUADALINFO_CONF)

            for line in fileobj.readlines():
                split_line = line.split('=')
                key = split_line[0].strip()
                value = split_line[1].strip()
                value = value.strip('"')
                #self[key] = value
                dict.__setitem__(self, key, value)

            fileobj.close()

        except IOError:
            self['ID_PROVINCIA'] = "0"
            self['ID_LOCALIDAD'] = "0"

    def __save(self):
        fileobj = open(GUADALINFO_CONF, 'w')
        for key, value in self.items():
            fileobj.write("%s=\"%s\"\n" % (key, value))
        fileobj.close()

    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
        self.__save()

    def __delitem__(self, key):
        dict.__delitem__(self, key)
        self.__save()



def parse_dpkg_call():
    result = Set()
    content = os.popen('COLUMNS=250 dpkg -l')
    for line in content:
        if line.startswith('ii'):
            name, version = line.split()[1:3]
            name = name.strip()
            version = version.strip()
            result.add((name,version))

            # Packages badformed
            try:
                extra_packages =  line.split(' ii ')[1:]
                for package in extra_packages:
                    name, version  = package.split()[0:2]
                    name = name.strip()
                    version = version.strip()
                    result.add((name, version))
            except IndexError:
                pass

    return result



# cicled redundancy !!
from managers import IncidenceManager
from managers import DataManager
from incidences import CentreCodeIncidence
